# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from datalad.api import clone
from datalad_next.exceptions import IncompleteResultsError
from datalad_next.tests.utils import (
    assert_in,
    assert_raises,
    eq_,
    skip_if_on_windows,
)
import datalad_osf.utils as dlosf_utils


def no_credentials(*args, **kwargs):
    return dict(token=None, username=None, password=None)


def test_readonly_access(tmp_path, monkeypatch):
    # obtain a prepared minimal dataset with pre-configured
    # OSF remotes and prestaged data
    # make sure that even with locally configured credentials
    # none actually reach the special remote
    with monkeypatch.context() as m:
        m.setattr(dlosf_utils, 'get_credentials', no_credentials)
        ds = clone(
            'https://github.com/datalad/testrepo--minimalds-osf.git',
            tmp_path,
        )
    # check that both OSF remotes were enabled
    assert_in('osfannex', ds.repo.get_remotes())
    assert_in('osftree', ds.repo.get_remotes())
    test_file = ds.repo.pathobj / 'inannex' / 'animated.gif'
    test_file_status = ds.repo.annexstatus([test_file])[test_file]
    eq_(test_file_status['has_content'], False)
    # git-annex reports annexed file to be available on osf remote
    eq_(ds.repo.get_special_remotes()[
        '82d91e46-58da-4625-96c0-56aadef98d49']['name'],
        'osfannex')
    assert_in(
        '82d91e46-58da-4625-96c0-56aadef98d49',
        ds.repo.whereis('inannex'))
    # obtain content, from osfannex specifically to avoid 'wget'
    ds.repo.call_git(['annex', 'copy', str(test_file), '-f', 'osfannex'])
    eq_(ds.repo.annexstatus([test_file])[test_file]['has_content'], True)


# git remote helper does not work on windows, due to some unclear
# line-ending(?) issue
# https://github.com/datalad/datalad-osf/pull/106#issuecomment-653772696
@skip_if_on_windows
def test_readonly_dataset_access(tmp_path, monkeypatch):
    # clone from OSF; ds is self-contained at OSF
    # make sure that even with locally configured credentials
    # none actually reach the special remote
    with monkeypatch.context() as m:
        m.setattr(dlosf_utils, 'get_credentials', no_credentials)
        ds = clone('osf://q8xnk', tmp_path)
    # standard name storage remote
    assert_in('osf-storage', ds.repo.get_remotes())
    for avail in ds.repo.whereis('inannex'):
        assert_in('7784367b-69c6-483d-9564-67f840715890', avail)
    test_file = ds.repo.pathobj / 'inannex' / 'animated.gif'
    eq_(ds.repo.annexstatus([test_file])[test_file]['has_content'], False)
    ds.repo.call_git(['annex', 'copy', str(test_file), '-f', 'osf-storage'])
    eq_(ds.repo.annexstatus([test_file])[test_file]['has_content'], True)


def test_invalid_url(tmp_path, monkeypatch):
    # make sure that even with locally configured credentials
    # none actually reach the special remote
    with monkeypatch.context() as m:
        m.setattr(dlosf_utils, 'get_credentials', no_credentials)
        assert_raises(
            IncompleteResultsError,
            clone, 'osf://q8xnk/somepath', tmp_path,
        )



#!/usr/bin/env python
## emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
"""

import sys
from urllib.parse import urlparse

from datalad_next.gitremotes.datalad_annex import (
    RepoAnnexGitRemote,
    main as datalad_annex_gitremote_main,
)


class OSFGitRemote(RepoAnnexGitRemote):
    """Very thin wrapper around ``RepoAnnexGitRemote``

    All this class does is to parse ``osf://...`` URLs and assemble the
    corresponding ``datalad-annex::`` URLs.

    Use of this class can be bypassed safely, with no loss of functionality.
    """
    def __init__(self,
                 gitdir,
                 remote,
                 url,
                 instream=sys.stdin,
                 outstream=sys.stdout,
                 errstream=sys.stderr):
        """
        All parameters are passed on to ``RepoAnnexGitRemote``.
        """
        self.parsed_url = urlparse(url)
        if self.parsed_url.path and self.parsed_url.path != '/':
            # project urls have no path
            raise RuntimeError("Only URLs of type osf://<PROJECT ID> are "
                               "supported. Got: %s" % url)

        osf_nodeid = self.parsed_url.netloc.strip('/')

        # we construct a datalad-annex:: URL and init the base class with it.
        # this will not expose all features of datalad-annex::, but users
        # requiring that could just switch to that directly. This is just a
        # backward compatibility/transition shim
        url = f'datalad-annex::?type=external&externaltype=osf&encryption=none&node={osf_nodeid}'

        super().__init__(
            gitdir=gitdir,
            remote=remote,
            url=url,
            instream=instream,
            outstream=outstream,
            errstream=errstream,
        )


def main():
    """git-remote helper executable entrypoint"""
    datalad_annex_gitremote_main(OSFGitRemote)


if __name__ == '__main__':
    main()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from datalad.interface.base import (
    Interface,
    build_doc,
)
from datalad.dochelpers import exc_str
from datalad.support.param import Parameter
from datalad.distribution.dataset import (
    datasetmethod,
    EnsureDataset,
)
from datalad.interface.base import (
    eval_results,
)
from datalad.support.constraints import (
    EnsureChoice,
    EnsureNone,
)
from datalad.downloaders.credentials import (
    Token,
    UserPassword,
)
from osfclient import OSF
from osfclient.exceptions import UnauthorizedException


@build_doc
class OSFCredentials(Interface):
    """Gather OSF credentials for subsequent non-interactive use

    This command enables (re-)entry of OSF credentials for storage in
    a credential manager. Once credentials are known, they will be
    retrieved automatically on demand, and enable non-interactive use
    for the purpose of data transfer to and from OSF.

    Credentials will be verified to enable successful authentication
    before being stored.
    """
    _params_ = dict(
        method=Parameter(
            args=("--method",),
            doc="""authentication method to use. 'token' authentication is
            strongly recommended.""",
            constraints=EnsureChoice("token", "userpassword")
        ),
        reset=Parameter(
            args=("--reset",),
            doc="""reset existing credentials and force re-entry""",
            action='store_true',
        ),
    )

    @staticmethod
    @datasetmethod(name='osf_credentials')
    @eval_results
    def __call__(method="token", reset=False):
        auth = None
        cred_spec = []
        if method == 'token':
            cred_spec = dict(token='token')
            auth = Token(
                name='https://osf.io',
                url='https://osf.io/settings/tokens',
            )
        elif method == 'userpassword':
            cred_spec = dict(user='username', password='password')
            auth = UserPassword(
                name='https://osf.io',
                url='https://osf.io/settings/account',
            )
        else:
            raise ValueError(
                'Unknown authentication method: {}'.format(method))
        if reset and auth.is_known:
            auth.delete()
        cred = {v: auth().get(k, None) for k, v in cred_spec.items()}

        # now verify that the credentials work by querying the
        # logged in user
        osf = OSF(**cred)
        try:
            req = osf.session.get('https://api.osf.io/v2/users/me/')
            req.raise_for_status()
        except UnauthorizedException:
            auth.delete()
            yield dict(
                action='osf_credentials',
                status='error',
                message='Invalid credentials',
                path=None,
            )
            return
        except Exception as e:
            yield dict(
                action='osf_credentials',
                status='impossible',
                message='Could not verify credentials, '
                        'please try again: {}'.format(
                            exc_str(e)),
                # needed to pacify DataLad 0.13.0 and earlier
                path=None,
            )
            return
        # if we get here auth has worked fine
        # get some attributes for an informative message
        attrs = req.json().get('data', {}).get('attributes', {})
        yield dict(
            action='osf_credentials',
            status='ok',
            message='authenticated{}{}{}'.format(
                ' as '
                if any(attrs.get(k, None) for k in ('email', 'full_name'))
                else '',
                attrs.get('full_name', ''),
                ' <{}>'.format(attrs['email'])
                if attrs.get('email', None)
                else ''),
            # needed to pacify DataLad 0.13.0 and earlier
            path=None,
            # report effective credentials
            **cred,
        )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from datalad.utils import (
    optional_args,
    wraps
)
from datalad_osf.utils import (
    create_node,
    delete_node,
    get_credentials,
)
from osfclient import OSF


@optional_args
def with_node(f, osf_session=None, title=None, category="data"):
    # we don't want the test hanging, no interaction
    creds = get_credentials(allow_interactive=False)
    # supply all credentials, so osfclient can fall back on user/pass
    # if needed
    osf = OSF(**creds)

    @wraps(f)
    def new_func(*args, **kwargs):
        node_id, proj_url = create_node(
            osf.session,
            'Temporary DataLad CI project: {}'.format(title),
            category=category)
        try:
            return f(*(args + (node_id,)), **kwargs)
        finally:
            delete_node(osf.session, node_id)

    return new_func



#!/usr/bin/env python3
# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##


import os
import json
from os.path import (
    dirname,
    basename,
)
import posixpath # OSF uses posix paths!
from urllib.parse import urlparse

from osfclient import OSF
from osfclient.exceptions import UnauthorizedException

from annexremote import (
    Master,
    ExportRemote,
    RemoteError,
)


class OSFSpecialRemote(ExportRemote):
    """git-annex special remote for the open science framework

    Any OSF node can be used as a remote, but the
    recommended setup is to create a subcomponent of your project
    for archiving your data. Mark it with the Data category so you
    can find it quickly and take note of its URL. Each component
    (or project) has a URL like https://osf.io/6zbyf/ which is
    needed to connect to it.

    .. todo::

       Write doc section on how to do that, and what URL should be used to
       configure the special remote.

    Initialize the special remote::

       git annex initremote osf type=external externaltype=osf \\
            encryption=none node=https://osf.io/<your-component-id>/

    To upload files you need to supply credentials.

    .. todo::

       Outline how this can be done

       OSF_USERNAME, OSF_PASSWORD - credentials, OR
       OSF_TOKEN   # TODO: untested, possibly currently broken in osfclient.

    Because this is a special remote, the uploaded data do not retain the
    git folder structure.

    The following parameters can be given to `git-annex initremote`, or
    `git annex enableremote` to (re-)configure a special remote.

    `node`
       the OSF URL of the file store

    `path`
       a subpath with (default: /)

    .. seealso::

        https://git-annex.branchable.com/special_remotes
          Documentation on git-annex special remotes

        https://osf.io
          Open Science Framework, a science-focused filesharing and archiving
          site.
    """
    def __init__(self, *args):
        super().__init__(*args)
        self.configs['node'] = 'The OSF URL for the remote'

        self.node = None

        # lazily evaluated cache of File objects
        self._files = None

    def initremote(self):
        ""
        if self.annex.getconfig('node') is None:
            raise ValueError('node URL or ID must be specified')
            # TODO: type-check the value; it must be https://osf.io/

    def prepare(self):
        """"""
        node_id = self.annex.getconfig('node')
        if not node_id:
            # fall back on outdated 'project' parameter, which could be
            # just the node ID or a full URL to a project
            node_id = posixpath.basename(
                urlparse(self.annex.getconfig('project')
                         ).path.strip(posixpath.sep))

        if not node_id:
            raise RemoteError('Could not determine OSF node ID')

        try:
            # make use of DataLad's credential manager for a more convenient
            # out-of-the-box behavior
            from datalad_osf.utils import get_credentials
            # we must stay non-interactive, because this is running inside
            # git-annex's special remote protocal
            creds = get_credentials(allow_interactive=False)
        except ImportError as e:
            # whenever anything goes wrong here, stay clam and fall back
            # on envvars.
            # we want this special remote to be fully functional without
            # datalad
            creds = dict(
                username=os.environ.get('OSF_USERNAME', None),
                password=os.environ.get('OSF_PASSWORD', None),
                token=os.environ.get('OSF_TOKEN', None),
            )
        # next one just sets up the stage, no requests performed yet, hence
        # no error checking needed
        # supply both auth credentials, so osfclient can fall back on user/pass
        # if needed
        osf = OSF(**creds)
        # next one performs initial auth
        try:
            self.node = osf.project(node_id)
        except Exception as e:
            # we need to raise RemoteError() such that PREPARE-FAILURE
            # is reported, sadly that doesn't give users any clue
            # TODO support datalad logging here
            raise RemoteError(
                'Failed to obtain OSF node handle: {}'.format(e)
            )
        # which storage to use, defaults to 'osfstorage'
        # TODO a node could have more than one? Make parameter to select?
        self.storage = self.node.storage()

    def transfer_store(self, key, filename):
        ""
        try:
            with open(filename, 'rb') as fp:
                self.storage.create_file(key, fp, force=True, update=True)
        except Exception as e:
            raise RemoteError(e)
        # we need to register the idea that this key is now present, but
        # we also want to avoid (re)requesting file info
        if self._files is not None:
            # assign None to indicate that we know this key, but
            # have no info from OSF about it
            self._files[key] = None

    def transferexport_store(self, key, local_file, remote_file):
        "Store the file located at `local_file` to `remote_file` on the remote"
        return self.transfer_store(remote_file, local_file)

    def transfer_retrieve(self, key, filename):
        """Get a key from OSF and store it to `filename`"""
        # we have to discover the file handle
        try:
            fobj = self.files[key]
            if fobj is None:
                # we have no info about this particular key -> trigger request
                self._files = None
                fobj = self.files[key]
            with open(filename, 'wb') as fp:
                fobj.write_to(fp)
        except Exception as e:
            # e.g. if the file couldn't be retrieved
            if isinstance(e, UnauthorizedException):
                # UnauthorizedException doesn't give a meaningful str()
                raise RemoteError('Unauthorized access')
            else:
                raise RemoteError(e)

    def transferexport_retrieve(self, key, local_file, remote_file):
        """Get the file located at `remote_file` from the remote

        and store it to `local_file`
        """
        return self.transfer_retrieve(remote_file, local_file)

    def checkpresent(self, key):
        "Report whether the OSF node has a particular key"
        try:
            if key not in self.files:
                # we don't know this key at all
                return False
            fobj = self.files.get(key, None)
            if fobj is None:
                # we knew the key, but never checked with OSF if it really
                # has it -> trigger request
                self._files = None
            return key in self.files
        except Exception as e:
            # e.g. if the presence of the key couldn't be determined, eg. in
            # case of connection error
            raise RemoteError(e)

    def checkpresentexport(self, key, remote_file):
        """Return  if the file `remote_file` is present in the remote"""
        return self.checkpresent(remote_file)

    def remove(self, key):
        """Remove a key from the remote"""
        f = self.files.get(key, None)
        if f is None:
            # removing a not existing key isn't considered an error
            return
        try:
            if f is None:
                self._files = None
                f = self.files[key]
            f.remove()
        except Exception as e:
            raise RemoteError(e)
        # anticipate change in remote and discard obj
        del self.files[key]

    def removeexport(self, key, remote_file):
        """Remove the file in `remote_file` from the remote"""
        return self.remove(remote_file)

    def _osf_makedirs(self, folder, path, exist_ok=False):
        """
        Create folder 'path' inside OSF folder object 'folder'.

        'folder' may also be a osfclient.models.storage.Storage,
          representing the root of a virtual OSF drive.

        Returns the final created folder object.
        """
        for name in path.strip(posixpath.sep).split(posixpath.sep):
            folder = folder.create_folder(name, exist_ok=exist_ok)

        return folder

    @property
    def files(self):
        if self._files is None:
            # get all file info at once
            # per-request latency is substantial, presumably it is overall
            # faster to get all at once
            self._files = {
                # strip leading prefix to be directly indexable with ani
                # annex key
                f.path.lstrip(posixpath.sep): f
                for f in self.storage.files
            }
            for p, key in (
                ('.git/refs', 'XDLRA--refs'),
                ('.git/repo.zip', 'XDLRA--repo-export'),
            ):
                handle = self._files.get(p)
                if handle is not None:
                    self._files[key] = handle
        return self._files

    def removeexportdirectory(self, remote_directory):
        """Remove the directory `remote_directory` from the remote"""
        try:
            folder = [f for f in self.storage.folders
                      if '{sep}{path}{sep}'.format(
                          sep=posixpath.sep,
                          path=remote_directory)]
            if not folder:
                # note that removing a not existing directory isn't
                # considered an error
                return
            elif len(folder) > 1:
                raise RuntimeError("More than matching folder found")
            folder = folder[0]
            # osfclient has no way to do this with the public API
            # going through the backdoor...
            folder._delete(folder._delete_url)
            # TODO delete all matching records from self._files
        except Exception as e:
            raise RemoteError(e)

    def renameexport(self, key, filename, new_filename):
        """Move the remote file in `name` to `new_name`"""
        try:
            fobj = self.files[filename]
            if fobj is None:
                # we have no info about this particular key -> trigger request
                self._files = None
                fobj = self.files[filename]
            response = self.storage.session.post(
                fobj._move_url,
                data=json.dumps(
                    dict(action='move',
                         path='/{}'.format(dirname(new_filename)),
                         rename=basename(new_filename))
                ),
            )
            if response.status_code != 201:
                raise RuntimeError('{}: {}'.format(response, response.text))
            del self._files[filename]
            self._files[new_filename] = None
        except Exception as e:
            raise RemoteError(repr(e))


def main():
    master = Master()
    remote = OSFSpecialRemote(master)
    master.LinkRemote(remote)
    master.Listen()


if __name__ == "__main__":
    main()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

"""DataLad extension for data exchange on the Open Science Framework"""

__docformat__ = 'restructuredtext'



# defines a datalad command suite
# this symbol must be identified as a setuptools entrypoint
# to be found by datalad
command_suite = (
    # description of the command suite, displayed in cmdline help
    "DataLad extension for OSF support",
    [
        ('datalad_osf.credentials', 'OSFCredentials',
            'osf-credentials', 'osf_credentials'),
        # specification of a command, any number of commands can be defined
        (
            # importable module that contains the command implementation
            'datalad_osf.create_sibling_osf',
            # name of the command class implementation in above module
            'CreateSiblingOSF',
            # optional name of the command in the cmdline API
            'create-sibling-osf',
            # optional name of the command in the Python API
            'create_sibling_osf'
        ),
    ]
)


from . import _version
__version__ = _version.get_versions()['version']



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from datalad_next.tests.utils import skip_if_on_windows

common_init_opts = ["encryption=none", "type=external", "externaltype=osf",
                    "autoenable=true"]


def test_gitannex(osf_node, minimal_dataset):
    from datalad.cmd import GitWitlessRunner
    ds = minimal_dataset

    # add remote parameters here
    init_remote_opts = ["node={}".format(osf_node)]

    # add special remote
    init_opts = common_init_opts + init_remote_opts
    ds.repo.init_remote('osfproject', options=init_opts)

    # run git-annex-testremote
    # note, that we don't want to capture output. If something goes wrong we
    # want to see it in test build's output log.
    # TODO use AnnexRepo._call_annex(..., protocol=None) with 0.14+
    GitWitlessRunner(
        cwd=ds.path,
        env=GitWitlessRunner.get_git_environ_adjusted()).run(
            ['git', 'annex', 'testremote', 'osfproject', "--fast"]
    )



import pytest


@pytest.fixture(autouse=False, scope="function")
def minimal_dataset(existing_dataset):
    ds = existing_dataset
    (ds.pathobj / 'file1.txt').write_text('content')
    (ds.pathobj / 'subdir').mkdir()
    (ds.pathobj / 'subdir' / 'file2.txt').write_text('different content')
    ds.save()

    yield ds


@pytest.fixture(autouse=False, scope="session")
def osf_credentials():
    """Yields credential dict from get_credentials() suitable of OSF client"""
    from datalad_osf.utils import get_credentials
    cred = get_credentials(allow_interactive=False)
    yield cred


@pytest.fixture(autouse=False, scope="function")
def osf_credentials_or_skip(osf_credentials):
    if not any(osf_credentials.values()):
        pytest.skip(reason='no OSF credentials')

    yield osf_credentials


@pytest.fixture(autouse=False, scope="function")
def osf_node(osf_credentials_or_skip):
    from datalad_osf.utils import (
        create_node,
        delete_node,
    )
    from osfclient import OSF
    osf = OSF(**osf_credentials_or_skip)

    title = 'Temporary DataLad CI project'
    category = "data"

    node_id, proj_url = create_node(
        osf.session, title, category=category,
    )
    yield node_id

    delete_node(osf.session, node_id)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import logging
from datalad.interface.base import (
    Interface,
    build_doc,
)
from datalad.support.annexrepo import AnnexRepo
from datalad.support.param import Parameter
from datalad.distribution.dataset import (
    datasetmethod,
    EnsureDataset,
    require_dataset
)
from datalad.interface.base import eval_results
from datalad.support.constraints import (
    EnsureChoice,
    EnsureNone,
    EnsureStr,
    EnsureBool,
)
from datalad.interface.results import get_status_dict
from datalad.interface.utils import ac
from osfclient import OSF
from datalad_osf.utils import (
    create_node,
    update_node,
    get_credentials,
)
from datalad.utils import ensure_list
from datalad.log import log_progress

lgr = logging.getLogger('datalad.osf.create_sibling_osf')


@build_doc
class CreateSiblingOSF(Interface):
    """Create a dataset representation at OSF.

    This will create a node on OSF and initialize
    an osf special remote to point to it. There are two modes
    this can operate in: 'annex' and 'export'.
    The former uses the OSF node as a key-value store, that
    can be used by git-annex to copy data to and retrieve
    data from (potentially by any clone of the original dataset).
    The latter allows to use 'git annex export' to publish a
    snapshot of a particular version of the dataset. Such an OSF
    node will - in opposition to the 'annex' - be
    human-readable.

    For authentication with OSF, you can define environment variables: Either
    'OSF_TOKEN', or both 'OSF_USERNAME' and 'OSF_PASSWORD'. If neither of these
    is defined, the tool will fall back to the datalad credential manager and
    inquire for credentials interactively.

    """

    result_renderer = 'tailored'

    _params_ = dict(
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc="""Dataset to create a sibling for.""",
            constraints=EnsureDataset() | EnsureNone()
        ),
        title=Parameter(
            args=("--title",),
            doc="""title of the to-be created OSF node that is displayed
            on the OSF website. Defaults to the basename of the root directory
            of the local dataset.""",
            constraints=EnsureStr() | EnsureNone(),
        ),
        name=Parameter(
            args=("-s", "--name",),
            doc="""Name of the to-be initialized osf-special-remote""",
            constraints=EnsureStr()
        ),
        storage_name=Parameter(
            args=("--storage-name",),
            metavar="NAME",
            doc="""Name of the storage sibling (git-annex special remote).
            Must not be identical to the sibling name. If not specified,
            defaults to the sibling name plus '-storage' suffix.""",
            constraints=EnsureStr() | EnsureNone()),
        existing=Parameter(
            args=("--existing",),
            constraints=EnsureChoice(
                'skip', 'error') | EnsureNone(),
            metavar='MODE',
            doc="""Action to perform, if a (storage) sibling is already
            configured under the given name and/or a target already exists.
            In this case, a dataset can be skipped ('skip'), or the command
            be instructed to fail ('error').""", ),
        trust_level=Parameter(
            args=("--trust-level",),
            metavar="TRUST-LEVEL",
            constraints=EnsureChoice(
                'trust', 'semitrust', 'untrust') | EnsureNone(),
            doc="""specify a trust level for the storage sibling. If not
            specified, the default git-annex trust level is used.""",),
        mode=Parameter(
            args=("--mode",),
            doc=""" """,
            constraints=EnsureChoice(
                "annex", "export", "exportonly", "gitonly")
        ),
        tags=Parameter(
            args=('--tag',),
            dest='tags',
            metavar='TAG',
            doc="""specific one or more tags for the to-be-create OSF node.
            A tag 'DataLad dataset' and the dataset ID (if there is any)
            will be automatically added as additional tags.
            [CMD: This option can be given more than once CMD].""",
            action='append',
        ),
        public=Parameter(
            args=("--public",),
            doc="""make OSF node public""",
            action='store_true',
        ),
        category=Parameter(
            args=("--category",),
            doc="""specific the OSF node category to be used for the
            node. The categorization determines what icon is displayed
            with the node on the OSF, and helps with search organization""",
            # all presently supported categories
            constraints=EnsureChoice(
                "analysis", "communication", "data", "hypothesis",
                "instrumentation", "methods and measures", "procedure",
                "project", "software", "other")
        ),
        description=Parameter(
            args=("--description",),
            metavar="TEXT",
            doc="""Description of the OSF node that will be displayed on
            the associated project page. By default a description will be
            generated based on the mode the sibling is put into.""",
            constraints=EnsureStr() | EnsureNone()),
    )

    @staticmethod
    @datasetmethod(name='create_sibling_osf')
    @eval_results
    def __call__(title=None,
                 name="osf",
                 storage_name=None,
                 dataset=None,
                 mode="annex",
                 existing='error',
                 trust_level=None,
                 tags=None,
                 public=False,
                 category='data',
                 description=None,
                 ):
        ds = require_dataset(dataset,
                             purpose="create OSF remote",
                             check_installed=True)
        res_kwargs = dict(
            ds=ds,
            action="create-sibling-osf",
            logger=lgr,
        )
        # we need an annex
        if not isinstance(ds.repo, AnnexRepo):
            yield get_status_dict(
                type="dataset",
                status="impossible",
                message="dataset has no annex",
                **res_kwargs)
            return

        # NOTES:
        # - we prob. should check osf-special-remote availability upfront to
        #   fail early
        # - add --recursive option
        #       - recursive won't work easily. Need to think that through.
        #       - would need a naming scheme for subdatasets
        #       - flat on OSF or a tree?
        #       - how do we detect something is there already, so we can skip
        #         rather than duplicate (with a new name)?
        #         osf-type-special-remote sufficient to decide it's not needed?
        # - adapt to conclusions in issue #30
        #   -> create those subcomponents
        # - results need to report URL for created projects suitable for datalad
        #   output formatting!
        #   -> result_renderer
        #   -> needs to ne returned by create_node

        if not storage_name:
            storage_name = "{}-storage".format(name)

        sibling_conflicts = sibling_exists(
            ds, [name, storage_name],
            # TODO pass through
            recursive=False, recursion_limit=None,
            # fail fast, if error is desired
            exhaustive=existing == 'error',
        )
        if existing == 'error' and sibling_conflicts:
            # we only asked for one
            conflict = sibling_conflicts[0]
            yield get_status_dict(
                status='error',
                message=(
                    "a sibling '%s' is already configured in dataset %s",
                    conflict[1], conflict[0]),
                **res_kwargs,
            )
            return

        if title is None:
            # use dataset root basename
            title = ds.pathobj.name

        tags = ensure_list(tags)
        if 'DataLad dataset' not in tags:
            tags.append('DataLad dataset')
        if ds.id and ds.id not in tags:
            tags.append(ds.id)

        if not description:
            description = \
                "This component was built from a DataLad dataset using the " \
                "datalad-osf extension " \
                "(https://github.com/datalad/datalad-osf)."
            if mode != 'exportonly':
                description += \
                    " With this extension installed, this component can be " \
                    "git or datalad cloned from a 'osf://ID' URL, where " \
                    "'ID' is the OSF node ID that shown in the OSF HTTP " \
                    "URL, e.g. https://osf.io/q8xnk can be cloned from " \
                    "osf://q8xnk. "
        cred = get_credentials(allow_interactive=True)
        osf = OSF(**cred)
        node_id, node_url = create_node(
            osf_session=osf.session,
            title=title,
            category=category,
            tags=tags if tags else None,
            public=EnsureBool()(public),
            description=description,
        )
        if mode != 'gitonly':
            init_opts = ["encryption=none",
                         "type=external",
                         "externaltype=osf",
                         "autoenable=true",
                         "node={}".format(node_id)]

            if mode in ("export", "exportonly"):
                init_opts += ["exporttree=yes"]

            ds.repo.init_remote(storage_name, options=init_opts)
            if trust_level:
                ds.repo.call_git(['annex', trust_level, storage_name])

            yield get_status_dict(
                type="dataset",
                url=node_url,
                id=node_id,
                name=storage_name,
                status="ok",
                **res_kwargs
            )

        if mode == 'exportonly':
            return

        # append how to clone this specific dataset to the description
        description += "This particular project can be cloned using" \
                       " 'datalad clone osf://{}'".format(node_id)
        update_node(osf_session=osf.session,
                    id_=node_id,
                    description=description)

        ds.config.set(
            'remote.{}.annex-ignore'.format(name), 'true',
            where='local')
        yield from ds.siblings(
            # use configure, not add, to not trip over the config that
            # we just made
            action='configure',
            name=name,
            url='osf://{}'.format(node_id),
            fetch=False,
            publish_depends=storage_name if mode != 'gitonly' else None,
            recursive=False,
            result_renderer=None,
        )

    @staticmethod
    def custom_result_renderer(res, **kwargs):
        from datalad.ui import ui
        if res['action'] == "create-sibling-osf":
            msg = res.get('message', None)
            ui.message("{action}({status}): {url}{msg}".format(
                action=ac.color_word(res['action'], ac.BOLD),
                status=ac.color_status(res['status']),
                url=res.get('url', ''),
                msg=' [{}]'.format(msg[0] % msg[1:]
                                   if isinstance(msg, tuple)
                                   else res['message'])
                if msg else '')
            )
        elif res['action'] == "add-sibling-osf":
            ui.message("{action}({status})".format(
                action=ac.color_word(res['action'], ac.BOLD),
                status=ac.color_status(res['status']))
            )
        else:
            from datalad.interface.utils import default_result_renderer
            default_result_renderer(res)


# TODO this could was originally taken from create_sibling_ria() and
# subsequently turned into a standalone function to facilitate re-use by other
# `create_sibling()` implementations
def sibling_exists(ds, names, recursive=False,
                   recursion_limit=None, exhaustive=False):
    # in recursive mode this check could take a substantial amount of
    # time: employ a progress bar (or rather a counter, because we don't
    # know the total in advance
    pbar_id = 'sibling-exists-{}'.format(id(ds))
    if recursive:
        log_progress(
            lgr.info, pbar_id,
            'Start checking pre-existing sibling configuration %s', ds,
            label='Query siblings',
            unit=' Siblings',
        )
    conflicts = []
    for r in ds.siblings(result_renderer=None,
                         recursive=recursive,
                         recursion_limit=recursion_limit):
        if recursive:
            log_progress(
                lgr.info, pbar_id,
                'Discovered sibling %s in dataset at %s',
                r['name'], r['path'],
                update=1,
                increment=True)
        if not r['type'] == 'sibling' or r['status'] != 'ok':
            # this is an internal status query that has no consequences
            continue
        for name in names:
            if r['name'] == name:
                conflicts.append((r['path'], name))
                if not exhaustive:
                    return conflicts
    if recursive:
        log_progress(
            lgr.info, pbar_id,
            'Finished checking pre-existing sibling configuration %s', ds,
        )
    return conflicts



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import json
from os import environ
from datalad.downloaders.credentials import (
    Token,
    UserPassword,
)
from datalad import ui


# Note: This should ultimately go into osfclient
def create_node(osf_session, title, category="data", tags=None,
                public=False, parent=None, description=None):
    """ Create a node on OSF

    Parameters
    ----------
    title: str
        Title of the node
    category: str
        categorization changes how the node is displayed
        on OSF, but doesn't appear to have a "real" function
    tags: list of str
    public: bool
        whether to make the new node public
    parent: str, optional
        ID of an OSF parent node to create a child node for

    Returns
    -------
    str
        ID of the created node
    """

    if parent:
        # we have a parent, use its URL to create children
        url = osf_session.build_url('nodes', parent, 'children')
    else:
        url = osf_session.build_url('nodes')
    post_data = {"data":
                     {"type": "nodes",
                      "attributes":
                          {"title": title,
                           "category": category,
                           "public": public,
                           }
                      }
                 }
    if tags:
        post_data["data"]["attributes"]["tags"] = tags
    if description:
        post_data["data"]["attributes"]["description"] = description

    response = osf_session.post(url, data=json.dumps(post_data))
    # TODO: figure what errors to better deal with /
    #       create a better message from
    response.raise_for_status()

    # TODO: This should eventually return an `node` instance (see osfclient).
    #       Response contains all properties of the created node.
    node_id = response.json()['data']['id']

    # Note: Going for "html" URL here for reporting back to the user, since this
    #       what they would need to go to in order to proceed manually.
    #       There's also the flavor "self" instead, which is the node's
    #       API endpoint.
    proj_url = response.json()["data"]["links"]["html"]
    return node_id, proj_url


def update_node(osf_session, id_, category=None, tags=None, description=None):
    """Updates a node on the osf

    id_: str
        to be updated node ID
    """
    url = osf_session.build_url('nodes', id_)
    patch_data = {"data":
                     {"type": "nodes",
                      "id": id_,
                      "attributes":
                          {}
                      }
                 }
    if category:
        patch_data["data"]["attributes"]["category"] = category
    if tags:
        patch_data["data"]["attributes"]["tags"] = tags
    if description:
        patch_data["data"]["attributes"]["description"] = description

    response = osf_session.patch(url, data=json.dumps(patch_data))
    response.raise_for_status()


def delete_node(osf_session, id_):
    """ Delete a node on OSF

    Parameters
    ----------
    id_: str
        to be deleted node ID
    """

    url = osf_session.build_url('nodes', id_)
    response = osf_session.delete(url)
    response.raise_for_status()


def initialize_osf_remote(remote, node,
                          encryption="none", autoenable="true"):
    """Initialize special remote with a given node

    convenience wrapper for git-annex-initremote w/o datalad

    Parameters
    ----------
    remote: str
        name for the special remote
    node: str
        ID of the node/component to use
    encryption: str
        see git-annex-initremote; mandatory option;
    autoenable: str
        'true' or 'false'; tells git-annex to automatically enable the
        special remote on git-annex-init (particularly after a fresh git-clone
    """

    init_opts = ["type=external",
                 "externaltype=osf",
                 "encryption={}".format(encryption),
                 "autoenable={}".format(autoenable),
                 "node={}".format(node)]

    import subprocess
    subprocess.run(["git", "annex", "initremote", remote] + init_opts)


def get_credentials(allow_interactive=True):
    # prefer the environment
    if 'OSF_TOKEN' in environ or all(
            k in environ for k in ('OSF_USERNAME', 'OSF_PASSWORD')):
        return dict(
            token=environ.get('OSF_TOKEN', None),
            username=environ.get('OSF_USERNAME', None),
            password=environ.get('OSF_PASSWORD', None),
        )

    # fall back on DataLad credential manager
    token_auth = Token(
        name='https://osf.io',
        url='https://osf.io/settings/tokens',
    )
    up_auth = UserPassword(
        name='https://osf.io',
        url='https://osf.io/settings/account',
    )

    do_interactive = allow_interactive and ui.is_interactive()

    # get auth token, from environment, or from datalad credential store
    # if known-- we do not support first-time entry during a test run
    token = environ.get(
        'OSF_TOKEN',
        token_auth().get('token', None)
        if do_interactive or token_auth.is_known
        else None)
    username = None
    password = None
    if not token:
        # now same for user/password if there was no token
        username = environ.get(
            'OSF_USERNAME',
            up_auth().get('user', None)
            if do_interactive or up_auth.is_known
            else None)
        password = environ.get(
            'OSF_PASSWORD',
            up_auth().get('password', None)
            if do_interactive or up_auth.is_known
            else None)

    return dict(token=token, username=username, password=password)



from datalad.conftest import setup_package

from datalad_next.tests.fixtures import (
    # no test can leave global config modifications behind
    check_gitconfig_global,
    # no test can leave secrets behind
    check_plaintext_keyring,
    # function-scope config manager
    datalad_cfg,
    # function-scope, Dataset instance
    dataset,
    # function-scope, Dataset instance with underlying repository
    existing_dataset,
)

from datalad_osf.tests.fixtures import (
    # standard test dataset setup used throughout the datalad-osf tests
    minimal_dataset,
    osf_credentials,
    osf_credentials_or_skip,
    osf_node,
)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from datalad_next.tests.utils import assert_result_count
from datalad_osf.utils import delete_node
from osfclient import OSF


#def test_invalid_calls(minimal_dataset):
#
#    # - impossible w/o dataset
#    # - impossible w/o annex
#    # - mandatory arguments
#    raise SkipTest("TODO")


def test_create_osf_simple(osf_credentials_or_skip, minimal_dataset):

    ds = minimal_dataset

    file1 = ds.pathobj / "file1.txt"

    create_results = ds.create_sibling_osf(name="osf")

    assert_result_count(create_results, 2, status='ok')
    assert_result_count(
        create_results, 1,
        status='ok', type='dataset', name="osf-storage", path=ds.path)
    assert_result_count(
        create_results, 1,
        status='ok', type='sibling', name="osf", path=ds.path)

    # if we got here, we created something at OSF;
    # make sure, we clean up afterwards
    try:
        # special remote is configured:
        remote_log = ds.repo.call_git(['cat-file', 'blob',
                                       'git-annex:remote.log'])
        assert "node={}".format(create_results[0]['id']) in remote_log

        # copy files over
        ds.repo.copy_to('.', "osf-storage")
        whereis = ds.repo.whereis(str(file1))
        here = ds.config.get("annex.uuid")
        # files should be 'here' and on remote end:
        assert len(whereis) == 2
        assert here in whereis

        # drop content here
        ds.drop('.')
        whereis = ds.repo.whereis(str(file1))
        # now on remote end only
        assert len(whereis) == 1
        assert here not in whereis

        # and get content again from remote:
        ds.get('.')
        whereis = ds.repo.whereis(str(file1))
        assert len(whereis) == 2
        assert here in whereis
    finally:
        # clean remote end:
        osf = OSF(**osf_credentials_or_skip)
        delete_node(osf.session, create_results[0]['id'])


def test_create_osf_export(osf_credentials_or_skip, minimal_dataset):

    ds = minimal_dataset

    create_results = ds.create_sibling_osf(
        title="CI dl-create",
        # do not create a git-remote
        mode="exportonly")

    assert_result_count(
        create_results, 1,
        status='ok', type='dataset', name='osf-storage', path=ds.path)

    # if we got here, we created something at OSF;
    # make sure, we clean up afterwards
    try:

        # for now just run an export and make sure it doesn't fail
        ds.repo.call_git(['annex', 'export', 'HEAD', '--to', 'osf-storage'])

    finally:
        # clean remote end:
        osf = OSF(**osf_credentials_or_skip)
        delete_node(osf.session, create_results[0]['id'])


# def test_create_osf_existing(osf_credentials_or_skip):
#     raise SkipTest("TODO")



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##


import sys
from setuptools import setup
import versioneer

from _datalad_buildsupport.setup import (
    BuildManPage,
)

cmdclass = versioneer.get_cmdclass()
cmdclass.update(build_manpage=BuildManPage)

# Give setuptools a hint to complain if it's too old a version
# 30.3.0 allows us to put most metadata in setup.cfg
# Should match pyproject.toml
SETUP_REQUIRES = ['setuptools >= 30.3.0']
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ['wheel'] if 'bdist_wheel' in sys.argv else []

if __name__ == '__main__':
    setup(name='datalad_osf',
          version=versioneer.get_version(),
          cmdclass=cmdclass,
          setup_requires=SETUP_REQUIRES,
          entry_points={
              # 'datalad.extensions' is THE entrypoint inspected by the datalad API builders
              'datalad.extensions': [
                  # the label in front of '=' is the command suite label
                  # the entrypoint can point to any symbol of any name, as long it is
                  # valid datalad interface specification (see demo in this extensions
                  'osf=datalad_osf:command_suite',
              ],
              'console_scripts': [
                  'git-remote-osf=datalad_osf.git_remote:main',
                  'git-annex-remote-osf=datalad_osf.annex_remote:main',
              ],
          },
    )

